@@ -57,6 +57,11 @@ Q_DECLARE_LOGGING_CATEGORY(QT_BT_BLUEZ)
 
 static const QLatin1String agentPath("/qt/agent");
 
+inline uint qHash(const QBluetoothAddress &address)
+{
+    return ::qHash(address.toUInt64());
+}
+
 QBluetoothLocalDevice::QBluetoothLocalDevice(QObject *parent) :
     QObject(parent), d_ptr(new QBluetoothLocalDevicePrivate(this))
 {
@@ -68,7 +73,7 @@ QBluetoothLocalDevice::QBluetoothLocalDevice(const QBluetoothAddress &address, Q
 }
 
 QString QBluetoothLocalDevice::name() const
-{    
+{
     if (!d_ptr || !d_ptr->adapter)
         return QString();
 
@@ -157,6 +162,11 @@ QBluetoothLocalDevice::HostMode QBluetoothLocalDevice::hostMode() const
     return HostPoweredOff;
 }
 
+QList<QBluetoothAddress> QBluetoothLocalDevice::connectedDevices() const
+{
+    return d_ptr->connectedDevices();
+}
+
 QList<QBluetoothHostInfo> QBluetoothLocalDevice::allDevices()
 {
     QList<QBluetoothHostInfo> localDevices;
@@ -337,9 +347,19 @@ QBluetoothLocalDevice::Pairing QBluetoothLocalDevice::pairingStatus(const QBluet
 }
 
 QBluetoothLocalDevicePrivate::QBluetoothLocalDevicePrivate(QBluetoothLocalDevice *q, QBluetoothAddress address)
-    : adapter(0), agent(0), localAddress(address), pendingHostModeChange(-1), msgConnection(0), q_ptr(q)
+    : adapter(0), agent(0), localAddress(address), pendingHostModeChange(-1), msgConnection(0), q_ptr(q), connectedCached(false)
 {
     initializeAdapter();
+    connectDeviceChanges();
+}
+
+void QBluetoothLocalDevicePrivate::connectDeviceChanges()
+{
+    if (adapter) { //invalid QBluetoothLocalDevice due to wrong local adapter address
+        connect(adapter, SIGNAL(PropertyChanged(QString,QDBusVariant)), SLOT(PropertyChanged(QString,QDBusVariant)));
+        connect(adapter, SIGNAL(DeviceCreated(QDBusObjectPath)), SLOT(_q_deviceCreated(QDBusObjectPath)));
+        connect(adapter, SIGNAL(DeviceRemoved(QDBusObjectPath)), SLOT(_q_deviceRemoved(QDBusObjectPath)));
+    }
 }
 
 QBluetoothLocalDevicePrivate::~QBluetoothLocalDevicePrivate()
@@ -347,6 +367,7 @@ QBluetoothLocalDevicePrivate::~QBluetoothLocalDevicePrivate()
     delete msgConnection;
     delete adapter;
     delete agent;
+    qDeleteAll(devices);
 }
 
 void QBluetoothLocalDevicePrivate::initializeAdapter()
@@ -414,6 +435,106 @@ void QBluetoothLocalDevicePrivate::RequestConfirmation(const QDBusObjectPath &in
     return;
 }
 
+void QBluetoothLocalDevicePrivate::_q_deviceCreated(const QDBusObjectPath &device)
+{
+    OrgBluezDeviceInterface *deviceInterface =
+            new OrgBluezDeviceInterface(QLatin1String("org.bluez"), device.path(), QDBusConnection::systemBus());
+    connect(deviceInterface, SIGNAL(PropertyChanged(QString,QDBusVariant)), SLOT(_q_devicePropertyChanged(QString,QDBusVariant)));
+    devices << deviceInterface;
+    QDBusPendingReply<QVariantMap> properties = deviceInterface->asyncCall(QLatin1String("GetProperties"));
+
+    properties.waitForFinished();
+    if (!properties.isValid()) {
+        qCritical() << "Unable to get device properties from: " << device.path();
+        return;
+    }
+    const QBluetoothAddress address = QBluetoothAddress(properties.value().value(QLatin1String("Address")).toString());
+    const bool connected = properties.value().value(QLatin1String("Connected")).toBool();
+
+    if (connectedCached) {
+        if (connected)
+            connectedDevicesSet.insert(address);
+        else
+            connectedDevicesSet.remove(address);
+    }
+    if (connected)
+        emit q_ptr->deviceConnected(address);
+    else
+        emit q_ptr->deviceDisconnected(address);
+}
+
+void QBluetoothLocalDevicePrivate::_q_deviceRemoved(const QDBusObjectPath &device)
+{
+    foreach (OrgBluezDeviceInterface *deviceInterface, devices) {
+        if (deviceInterface->path() == device.path()) {
+            devices.remove(deviceInterface);
+            delete deviceInterface; //deviceDisconnected is already emitted by _q_devicePropertyChanged
+            break;
+        }
+    }
+}
+
+void QBluetoothLocalDevicePrivate::_q_devicePropertyChanged(const QString &property, const QDBusVariant &value)
+{
+    OrgBluezDeviceInterface *deviceInterface = qobject_cast<OrgBluezDeviceInterface*>(sender());
+    if (deviceInterface && property == QLatin1String("Connected")) {
+        QDBusPendingReply<QVariantMap> propertiesReply = deviceInterface->GetProperties();
+        propertiesReply.waitForFinished();
+        if (propertiesReply.isError()) {
+            qCWarning(QT_BT_BLUEZ) << propertiesReply.error().message();
+            return;
+        }
+        const QVariantMap properties = propertiesReply.value();
+        const QBluetoothAddress address = QBluetoothAddress(properties.value(QLatin1String("Address")).toString());
+        const bool connected = value.variant().toBool();
+
+        if (connectedCached) {
+            if (connected)
+                connectedDevicesSet.insert(address);
+            else
+                connectedDevicesSet.remove(address);
+        }
+        if (connected)
+            emit q_ptr->deviceConnected(address);
+        else
+            emit q_ptr->deviceDisconnected(address);
+    }
+}
+
+void QBluetoothLocalDevicePrivate::createCache()
+{
+    if (!adapter)
+        return;
+
+    QDBusPendingReply<QList<QDBusObjectPath> > reply = adapter->ListDevices();
+    reply.waitForFinished();
+    if (reply.isError()) {
+        qCWarning(QT_BT_BLUEZ) << reply.error().message();
+        return;
+    }
+    foreach (const QDBusObjectPath &device, reply.value()) {
+        OrgBluezDeviceInterface deviceInterface(QLatin1String("org.bluez"), device.path(), QDBusConnection::systemBus());
+
+        QDBusPendingReply<QVariantMap> properties = deviceInterface.asyncCall(QLatin1String("GetProperties"));
+        properties.waitForFinished();
+        if (!properties.isValid()) {
+            qCWarning(QT_BT_BLUEZ) << "Unable to get properties for device " << device.path();
+            return;
+        }
+
+        if (properties.value().value(QLatin1String("Connected")).toBool())
+            connectedDevicesSet.insert(QBluetoothAddress(properties.value().value(QLatin1String("Address")).toString()));
+    }
+    connectedCached = true;
+}
+
+QList<QBluetoothAddress> QBluetoothLocalDevicePrivate::connectedDevices() const
+{
+    if (!connectedCached)
+        const_cast<QBluetoothLocalDevicePrivate *>(this)->createCache();
+    return connectedDevicesSet.toList();
+}
+
 void QBluetoothLocalDevice::pairingConfirmation(bool confirmation)
 {
     if(!d_ptr ||
